<?php
require_once ('LineItem.php');
require_once ('EntryList.php');
require_once ('ProductEntry.php');
require_once ('dbutils.php');
/*
 * This class is used for reading in a file with the definition
 * of types (drink, meal, etc.) and products.
 * 
 * The file uses the space indenting for assignment to higher level items.
 * If an item has no deeper elements it is considered a product, not
 * a type.
 * 
 * A sample can look like
 * 
 * Type1
 * 		Product A
 * 		Product B
 * 		Type 3
 * 			Product C
 * 			Type 7
 * 				Product D
 * 				Product E
 * 			Type 8
 * 				Product F
 * 		Type 4
 * 			Product G
 * 
 * A type can have additional attributes after "=":
 * 	- nothing: do not run the products in this category over kitchen and supplydesk
 * 			(for example: waiter mixes everything in bar and delivers at same time)
 *  - B: do not run the products in this category, but over supplydesk
 *  		(for example: drinks are mixed by waiter at bar, not in kitchen)
 *  - KB: run all products in this category over kitchen and supplydesk
 *  		(the default, if not overwritten by "type = " in higher category)
 *  - K: over kitchen, not over supplydesk
 *  		(for example: cook delivers product himself)
 */

class TypeAndProductFileManager {
	
	private $entries;   // of type EntryList  (in the beginning all items, later the types)
	private $leafArray;  // later filled by all leafs (= products) when filtered out from $entries;
	
	private $nextIdOfProdType = 1;
	
	var $dbutils;
	
	function __construct() {
		$this->dbutils = new DbUtils();
	}
	
	/*
	 * Look at the beginning of a line and count the number of spaces or tabs
	 */
	private function intendingOfText($text) {
		$charCounter = 0;
		while (($text[$charCounter] == ' ') || ($text[$charCounter] == "\t")) {
			$charCounter++;
		}
		return $charCounter;
	}
	
	private function startsWith($aText, $needle)
	{
		return $needle === "" || strpos($aText, $needle) === 0;
	}
	
	/*
	 * read in the "Speisekarte.txt"
	 * put the content in the array list "EntryList" as it is
	 * without any modifications
	 */
	private function readFile($fileName) {
		// remove old content if any
		$this->entries = new EntryList();
		
		// first add a dummy so that id and index will equal in the beginning
		// (later ID in DB is also starting from 1)
		$this->entries->add(new LineItem(0,0,"Dummy"));  	// index 0
		
		// now really read the file so that content starts at index 1
		$index = 1;
		$handle = fopen ($fileName, "r");
		while (!feof($handle)) {
			$textline = fgets($handle);
			if(!($this->startsWith($textline,'#'))) {
				$depth = $this->intendingOfText($textline);
				$this->entries->add(new LineItem($depth,$index,$textline));
				$index++;
			}
		}
		fclose ($handle);
	}
	
	private function findNextIdOfProdType($pdo) {
		$index = 1;
		
		$sql = "SELECT id FROM %prodtype% ORDER BY id DESC LIMIT 1";
		$stmt = $pdo->prepare($this->dbutils->resolveTablenamesInSqlString($sql));
		$stmt->execute();
		$row =$stmt->fetchObject();
		if ($row != null) {
			$index = intval($row->id) +1;
		}
		$this->nextIdOfProdType = $index;
	}
	
	/*
	 * read in the "Speisekarte.txt"
	* put the content in the array list "EntryList" as it is
	* without any modifications
	*/
	private function parseContent($speisekarte) {
		// remove old content if any
		$this->entries = new EntryList();
	
		// get max number of prodtype id -> the old ones will be kept!
		$index = $this->nextIdOfProdType;
		
		$lines = explode("\n", $speisekarte);
		
		for ($i=0;$i<count($lines);$i++) {
			$textline = $lines[$i];
			$cleanLine = str_replace(" ", "", $textline);
			if(!($this->startsWith($textline,'#')) && (strlen($cleanLine) > 0)) {
				$depth = $this->intendingOfText($textline);  
				$this->entries->add(new LineItem($depth,$index,$textline));
				$index++;
			}
		}
	}
	
	/*
	 * if something is a leaf node it means that it has no
	 * more susequent items with a higher depth.
	 * 
	 * In the sense of the Speisekarte this is a 'product' and
	 * not a 'type'.
	 */
	private function isLeafNode($index) {
		if ($index == ($this->entries->size()-1)) {
			// last entry -> it must be a leaf
			return true;
		}
		
		$currentEntry = $this->entries->get($index);
		$nextEntry = $this->entries->get($index+1);
		$currentDepth = $currentEntry->getDepth();
		$nextDepth = $nextEntry->getDepth();
		
		if ($nextDepth > $currentDepth) {
			return false;
		} else {
			return true;
		}
	}
	
	/*
	 * This is like if the entry at index $abIndex would have
	 * been removed from the array. (In fact this will be done
	 * after call of this method probably).
	 * This means that all the entries later will have an index
	 * that has to be decremented by 1. In this case index=id
	 * (the index is stored in the id).
	 * If lower elements have references to the elements with 
	 * decreased indeces, they have to lower their reference as
	 * well.
	 */
	private function reduceIndex($abIndex) {
		// now decrease the index and references if the refer to the reduced part
		$theEntryToRemove = $this->entries->get($abIndex);
		$idOfEntryToRemove = intval($theEntryToRemove->getId());
	    for ($i=$abIndex;$i<$this->entries->size();$i++) {
	    	$entry = $this->entries->get($i);
	    	$theId = intval($entry->getId()); // maybe not necessary going by id
	    	$ref = intval($entry->getReference());
	    	$entry->setId($theId-1);
	    	$theName = $entry->getName();
	    	if ($ref >= $idOfEntryToRemove) {
	    		$entry->setReference($ref-1);
	    	}
	    }
	}
	
	/*
	 * Recursive called function!
	 * 
	 * Find all subitems (with a deeper indenting) and start this
	 * method on these deeper items.
	 * 
	 * This method sets the references, so that it has no return value.
	 */
	private function findAllSubItemsOfIndex($index,$useKitchen,$useSupplyDesk,$kind) {
		$currentEntry = $this->entries->get($index);
		$currentId = $currentEntry->getId(); // maybe currentId = $index enough?
		$currentDepth = $currentEntry->getDepth();

		// now set the type (leaf -> product, not lead -> type)
		if ($this->isLeafNode($index)) {
			$currentEntry->setType(ENTRY_PRODUCT);
		} else {
			$currentEntry->setType(ENTRY_TYPE);
			$useKitOfEntry = $currentEntry->getUseKitchen();
			$useSupplyOfEntry = $currentEntry->getUseSupplyDesk();
			$kindOfEntry = $currentEntry->getKind();
			
			if ($useKitOfEntry < 0) {
				$this->entries->get($index)->setUseKitchen($useKitchen);
			} else {
				$useKitchen = $useKitOfEntry;
			}
			if ($useSupplyOfEntry < 0) {
				$this->entries->get($index)->setUseSupplyDesk($useSupplyDesk);
			} else {
				$useSupplyDesk = $useSupplyOfEntry;
			}
			if ($kindOfEntry == KIND_UNDEFINED) {
				$this->entries->get($index)->setKind($kind);
			} else {
				$kind = $kindOfEntry;
			}
		}
		
		// now look at all the subsequent items after this item
		for ($nextIndex=($index+1);$nextIndex<$this->entries->size();$nextIndex++) {
			$nextEntry = $this->entries->get($nextIndex);
			$nextId = $nextEntry->getId();
			$nextDepth = $nextEntry->getDepth();
			
			if ($nextDepth == ($currentDepth + 1)) {
				// directly under currentDepth
				$nextEntry->setReference($currentId);
				$this->findAllSubItemsOfIndex($nextIndex,$useKitchen,$useSupplyDesk,$kind); // recursive!
			} else if ($nextDepth <= $currentDepth) {
				// same depth level or higher - is no sub-element
				break;
			}
		}
		
		
	}

	
	/*
	 * Sort out the leafs from the entries. In the entries only the 
	 * types shall be left. The leafs (products) shall go to another
	 * array.
	 */
	private function sortOutLeafs() {
		$this->leafArray = new EntryList(); // is of same type
		for ($index=0;$index < $this->entries->size();$index++) {
			$currentEntry = $this->entries->get($index);
			if ($currentEntry->getType() == ENTRY_PRODUCT) {
				$this->leafArray->add($currentEntry);
				// now modify the references from current index+1 by a ref-1
				$this->reduceIndex($index);
				
				$this->entries->remove($index);
				$index--;
			}
		}
	}
	
	/*
	 * Fill the table prodtype in the DB. The access is done via the 
	 * view name.
	 */
	private function fillProdTypeDbTable ($pdo) {
		for ($i = 0;$i < $this->entries->size(); $i++) {
			$usekitchen = 1; // default
			$usesupplydesk = 1; // default
			$theEntry = $this->entries->get($i);
			
			$parts = explode(';', $theEntry->getName(), 2);
			//$parts = explode(';', "123 ; 45", 2);
			$theNameEntry = $parts[0];
			$usekitchen = $theEntry->getUseKitchen();
			$usesupplydesk = $theEntry->getUseSupplyDesk();
			$kind = $theEntry->getKind();
			
			$theProdTypeName = trim($theNameEntry);
			$theRefId = $theEntry->getReference();
			//$id = $i + $this->nextIdOfProdType;
			$id = $theEntry->getId();
			
			$insertSql = "INSERT INTO `%prodtype%` (`id`, `name`, `reference`, `usekitchen`, `usesupplydesk`, `kind`) VALUES (?,?,?,?,?,?)";
			$stmt = $pdo->prepare($this->dbutils->resolveTablenamesInSqlString($insertSql));
			if ($theEntry->getDepth() > 0) {
				$stmt->execute(array($id,$theProdTypeName,$theRefId,$usekitchen,$usesupplydesk,$kind));
			} else {
				$stmt->execute(array($id,$theProdTypeName,null,$usekitchen,$usesupplydesk,$kind));
			}
		}
	}
	
	private function fillProductDbTable($pdo,$leafArray) {
		for ($i=0;$i < $leafArray->size(); $i++) {
			$theLeafEntry = $leafArray->get($i);
			$product = new ProductEntry($theLeafEntry->getName());
			$priceA = $product->getPriceA();
			$priceB = $product->getPriceB();
			$priceC = $product->getPriceC();
			$shortName = $product->getShortName();
			$longName = $product->getLongName();
			$available = 1; // default: product is available
			$category = $theLeafEntry->getReference();
			
			$sql = "INSERT INTO `%products%` (`id`, `shortname`, `longname`,  `priceA`,  `priceB`,  `priceC`, `category`,`available`) VALUES (NULL,";
			$sql = $sql . "?,?,?,?,?,?,?)";
			$stmt = $pdo->prepare($this->dbutils->resolveTablenamesInSqlString($sql));
			$stmt->execute(array($shortName,$longName,$priceA,$priceB,$priceC,$category,$available));
		}
	}
	
	/*
	 * Only for test purposes: display entries (=types) and leafs
	 */
	private function outputAll() {
		echo "Types: <br>";
		echo "<table border=1>";
		echo "<tr><th>Index<th>Name<th>Referenz<th>K&uuml;che<th>Bereitstellung</tr>";
		for ($i=1;$i < $this->entries->size(); $i++) {
			$theEntry = $this->entries->get($i);
			if ($theEntry->getReference() > 0) {
				$theRefEntry = $this->entries->get($theEntry->getReference());
				echo "<tr><td>$i<td>" . $theEntry->getName() . "<td>" . $theRefEntry->getName() . "<td>" . $theEntry->getUseKitchen() . "<td>" . $theEntry->getUseSupplyDesk() . "</tr>";
				//echo $i . ":" . $theEntry->getName() . " -> " . $theRefEntry->getName() . "<br>";
			} else {
				echo "<tr><td>$i<td>" . $theEntry->getName() . "<td>null" . "<td>" . $theEntry->getUseKitchen() . "<td>" . $theEntry->getUseSupplyDesk() . "</tr>";
				//echo $i . ":" . $theEntry->getName() . " -> null<br>";
			}
		}
		echo "</table>";
		
		echo "<br>Leafs: <br>";
		for ($i=1;$i < $this->leafArray->size(); $i++) {
			$theEntry = $this->leafArray->get($i);
			if ($theEntry->getReference() > 0) {
				$theRefEntry = $this->entries->get($theEntry->getReference());
				$product = new ProductEntry($theEntry->getName());
				// echo $theEntry->getName() . " -> " . $theRefEntry->getName() . " ---- ";
				echo $product->toString() . " -> " . $theRefEntry->getName() . "<br>";
			} else {
				echo $theEntry->getName() . " -> null<br>";
			}
		}
	}

	
	function manageSpeisekarte($pdo,$content) {
		$this->findNextIdOfProdType($pdo);
		// first remove previous content
		$sql = "UPDATE %products% SET removed='1'";
		$stmt = $pdo->prepare($this->dbutils->resolveTablenamesInSqlString($sql));
		$stmt->execute();

		$sql = "UPDATE `%prodtype%` SET removed='1'";
		$stmt = $pdo->prepare($this->dbutils->resolveTablenamesInSqlString($sql));
		$stmt->execute();
		// now fill in the new content
		$this->parseContent($content);
	
		// now look on the first depth level (0). This level has
		// to be treated individually as staring point because later
		// on it will be iterated recursively
	
		// look at all elements at depth level 0
		
		for ($i=0;$i<$this->entries->size();$i++) {
			$anEntry = $this->entries->get($i);

			if ($anEntry->getDepth() == 0) {
				// highest level
				$this->findAllSubItemsOfIndex($i,1,1,FOOD);
			}
		}
		
		//Debug ausgabe
		for ($i=0;$i<$this->entries->size();$i++) {
			$anEntry = $this->entries->get($i);
			$theId = $anEntry->getId();
			$theName = $anEntry->getName();
			$theRef = $anEntry->getReference();
		}
		
		$this->sortOutLeafs();

		$this->fillProdTypeDbTable($pdo);
		$this->fillProductDbTable($pdo,$this->leafArray);
	}
}
?>